/* mld.c - Module Linker */

// Written by Byron Stanoszek  - gandalf@winds.org
// vim:tw=99
//
// Revision history:
//  1.0 - Sep 17, 2003: Initial revision
//  1.1 - May 31, 2004: Support 64-bit ELF files.

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <elf.h>
#include <sys/stat.h>
#include <sys/mman.h>

#define VERSION "1.1 - May 31 2004"

#define quad long long

#define ELF32  Elf32_Ehdr
#define ELF64  Elf64_Ehdr

#define SECT32 Elf32_Shdr
#define REL32  Elf32_Rel
#define RELA32 Elf32_Rela
#define SYM32  Elf32_Sym

#define SECT64 Elf64_Shdr
#define REL64  Elf64_Rel
#define RELA64 Elf64_Rela
#define SYM64  Elf64_Sym

#define Sym64(i)  ELF64_R_SYM(i)
#define Type64(i) ELF64_R_TYPE(i)

#define Sym32(i)  ELF32_R_SYM(i)
#define Type32(i) ELF32_R_TYPE(i)

/* Maximum section size: 128 MB */

/* 64-bit or 96-bit relocation record */
struct reloc {
  unsigned int offset:27;  // Source Section Offset
  unsigned int dest:2;     // Destination Section (0=Absolute, 1=.text, 2=.data, 3=.bss)
  unsigned int mode:2;     // Addressing Mode (0=32 Absolute, 1=32 PC-Relative, 2=64 Absolute)
  unsigned int src:1;      // Source Section (0=.text, 1=.data)

  unsigned quad value;     // Relocation Value
} __attribute__((__packed__));

/* Structure to hold list of undefined symbols */
struct symlist {
  struct symlist *next;
  char *name;
};

static int handle_32bit(ELF32 *head, unsigned char *input, int size);
static int handle_64bit(ELF64 *head, unsigned char *input, int size);
static int elf_header(ELF32 *elf, char *file, int size);

static char *file;
static char *outfile="a.out";
static int verbose;

/*************************************************************************************************/

int main(int argc, char *argv[])
{
  ELF32 *head;
  struct stat statbuf;
  unsigned char *input;
  char *cmd=*argv;
  int fd, c, size, elf64;

  while((c=getopt(argc, argv, "o:hvV?")) != -1) {
    switch(c) {
    case 'o':  /* Output File Name */
      outfile=optarg;
      break;
    case 'v':  /* Verbose Level */
      verbose++;
      break;
    case ':':
      fprintf(stderr, "%s: option requires an argument -- %c\n", cmd, optopt);
      return 1;
    case 'h':
    case '?':
    usage:
      printf("Usage: %s [-o outputfile] [-hvV] file.o\n", cmd);
      return 0;
    case 'V':  /* Display version information */
      printf("Module Linker version %s\n", VERSION);
      return 0;
    }
  }

  if(!argv[optind])
    goto usage;
  file=argv[optind];

  /* Open specified file */
  if((fd=open(file, O_RDONLY)) == -1) {
    perror(file);
    return 1;
  }

  /* Find total file size */
  if(lstat(file, &statbuf)) {
    perror(file);
    return 1;
  }
  if(statbuf.st_size <= sizeof(ELF64)) {
    fprintf(stderr, "%s: Input file too small\n", file);
    return 1;
  }
  size=statbuf.st_size;

  /* Map object file into memory */
  if(!(input=mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0))) {
    perror(file);
    return 1;
  }

  /* Validate ELF Header */
  head=(ELF32 *)input;
  elf64=elf_header(head, file, size);
  if(!elf64--)
    return 1;

  return elf64 ? handle_64bit((ELF64 *)head, input, size)
               : handle_32bit(head, input, size);
}

// Process a 32-bit ELF File.
//
static int handle_32bit(ELF32 *head, unsigned char *input, int size)
{
  SECT32 *sect;
  RELA32 *rel;
  SYM32 *sym=NULL;
  struct stat statbuf;
  struct reloc reloc;
  struct symlist *undef_list=NULL, *ptr, *last;
  int text=-1, data=-1, bss=-1;
  int seg, type, link, strlink, text_offset, addend;
  int i, j, n, outfd, error=0;
  char *strtab=NULL, *name;

  /* Validate section table */
  sect=(SECT32 *)(input+head->e_shoff);
  for(i=0;i < head->e_shnum;i++) {
    if(sect[i].sh_type != SHT_NOBITS) {
      if(sect[i].sh_offset > size || sect[i].sh_offset < 0 ||
         sect[i].sh_offset+sect[i].sh_size > size || sect[i].sh_offset+sect[i].sh_size < 0) {
        fprintf(stderr, "%s: Section %d has invalid size\n", file, i);
        return 1;
      } if(sect[i].sh_size >= 0x8000000) {
        fprintf(stderr, "%s: Section %d is larger than 128 MB\n", file, i);
        return 1;
      }
    }

    /* Determine file offset for String Table */
    if(!strtab && sect[i].sh_type == SHT_STRTAB)
      strtab=input+sect[i].sh_offset;
  }
  if(!strtab) {
    fprintf(stderr, "%s: No string table present\n", file);
    return 1;
  }

  /* Print sections */
  if(verbose) {
    printf("Sections in %s:\n\n", file);
    printf("Sect: Name                Offset    Size      Info Link\n");
    for(i=0;i < head->e_shnum;i++)
      if(sect[i].sh_type != SHT_NULL)
        printf("%4d: %-18.18s  %08x  %08x  %-4d %-4d\n", i, strtab+sect[i].sh_name,
               sect[i].sh_offset, sect[i].sh_size, sect[i].sh_info, sect[i].sh_link);
  }


  /* Open output file */
  if((outfd=open(outfile, O_WRONLY|O_CREAT|O_TRUNC, 0666)) == -1) {
    perror(outfile);
    return 1;
  }

  /* Write .text .data .bss segment sizes */
  for(i=0,seg=0;i < head->e_shnum;i++)
    if(!strcmp(strtab+sect[i].sh_name, ".text")) {
      text=i;
      seg=sect[i].sh_size;
      break;
    }
  if(i == head->e_shnum) {
    fprintf(stderr, "%s: Missing .text section\n", file);
    goto error;
  }
  write(outfd, &seg, sizeof(int));

  for(i=0,seg=0;i < head->e_shnum;i++)
    if(!strcmp(strtab+sect[i].sh_name, ".data")) {
      data=i;
      seg=sect[i].sh_size;
      break;
    }
  write(outfd, &seg, sizeof(int));

  for(i=0,seg=0;i < head->e_shnum;i++)
    if(!strcmp(strtab+sect[i].sh_name, ".bss")) {
      bss=i;
      seg=sect[i].sh_size;
      break;
    }
  write(outfd, &seg, sizeof(int));

  /* Segment text offset in output file (Write 0 for now, fill in later) */
  seg=0;
  write(outfd, &seg, sizeof(int));


  /* Process Symbol Relocations */
  for(i=0;i < head->e_shnum;i++) {
    if(sect[i].sh_type != SHT_RELA && sect[i].sh_type != SHT_REL)
      continue;
    if((seg=sect[i].sh_info) < 0 || seg >= head->e_shnum || sect[seg].sh_type != SHT_PROGBITS)
      continue;
    if((link=sect[i].sh_link) < 0 || link >= head->e_shnum || sect[link].sh_type != SHT_SYMTAB)
      continue;
    if((strlink=sect[link].sh_link) < 0 || strlink >= head->e_shnum ||
       sect[strlink].sh_type != SHT_STRTAB)
      continue;

    /* Loop on # symbols per section */
    for(n=0;n < sect[i].sh_size/sizeof(REL32);n++) {
      if(sect[i].sh_type == SHT_REL) {
        rel=(RELA32 *)((REL32 *)(input+sect[i].sh_offset)+n);
        addend=0;
      } else {
        rel=(RELA32 *)(input+sect[i].sh_offset)+n;
        addend=rel->r_addend;
      }
      type=Type32(rel->r_info);
      j=Sym32(rel->r_info);
      if((type != 1 && type != 2) || j < 0 || j >= sect[link].sh_size/sizeof(SYM32) ||
         rel->r_offset > sect[seg].sh_size) {
        fprintf(stderr, "%s: Invalid relocation type %d at Section %d, Entry %d\n", file,
                type, i, n);
        goto error;
      }

      /* Determine symbol name */
      sym=(SYM32 *)(input+sect[link].sh_offset)+j;
      if(sym->st_info == STT_SECTION)
        name=(sym->st_shndx < head->e_shnum) ?
             strtab+sect[sym->st_shndx].sh_name:"";
      else {
        if(sect[strlink].sh_offset+sym->st_name > size) {
          fprintf(stderr, "%s: Error retrieving symbol name for Section %d, Entry %d\n",
                  file, i, n);
          goto error;
        }
        name=input+sect[strlink].sh_offset+sym->st_name;
      }

      /* Display relocation list */
      if(verbose > 1) {
        if(!n)
          printf("\nSection %s relocations:\n", strtab+sect[seg].sh_name);
        printf("%4d: %-18.18s  %08x  %5s:%08x  %s\n", n, name, rel->r_offset, !sym->st_shndx?
               "UNDEF":sym->st_shndx < head->e_shnum ? strtab+sect[sym->st_shndx].sh_name:
               sym->st_shndx == SHN_ABS?"ABS":sym->st_shndx == SHN_COMMON?"COM":"?",
               sym->st_value+addend, type == 2?"PC Relative":""); 
      }

      /* Record a list of undefined symbols */
      if(!sym->st_shndx) {
        for(ptr=undef_list,last=NULL;ptr;last=ptr,ptr=ptr->next)
          if(!strcmp(ptr->name, name))
            break;
        if(!ptr) {
          ptr=malloc(sizeof(struct symlist));
          ptr->next=NULL;
          ptr->name=strdup(name);
          if(last)
            last->next=ptr;
          else
            undef_list=ptr;
        }
        continue;
      }

      /* If this is an absolute address that is known, modify segment directly */
      if(sym->st_shndx == SHN_ABS && type == 1) {
        *(int *)(input+sect[seg].sh_offset+rel->r_offset) += sym->st_value+addend;
        continue;
      }

      /* If this symbol is Common, delay error reporting */
      if(sym->st_shndx == SHN_COMMON) {
        error=1;
        continue;
      }

      /* Otherwise store in output module file */
      reloc.src=(seg == data);
      reloc.mode=(type == 2);
      reloc.dest=(sym->st_shndx == text)?1:(sym->st_shndx == data)?2:(sym->st_shndx == bss)?3:0;
      reloc.offset=rel->r_offset;
      reloc.value=sym->st_value+addend;
      write(outfd, &reloc, 8);
    }
  }

  if(verbose > 1)
    putchar('\n');

  /* Linker encountered an error */
  if(undef_list || error) {
    /* List undefined symbols */
    for(ptr=undef_list;ptr;ptr=ptr->next)
      fprintf(stderr, "%s: Error: symbol `%s' is undefined\n", file, ptr->name);

    /* Report use of Common variables */
    if(error) {
      fprintf(stderr, "%s: Error: At least one symbol was defined as Common.\n", file);
      fprintf(stderr, "%s: Error: Recompile the program using -fno-common.\n", file);
    }

    error:

    /* Remove temporary file on error */
    fstat(outfd, &statbuf);
    if(S_ISREG(statbuf.st_mode))
      unlink(outfile);

    close(outfd);
    return 1;
  }

  /* Write text and data segments to file */
  text_offset=lseek(outfd, 0, SEEK_CUR);
  if(write(outfd, input+sect[text].sh_offset, sect[text].sh_size) == -1) {
    perror(outfile);
    goto error;
  }
  if(data != -1 && sect[data].sh_size)
    if(write(outfd, input+sect[data].sh_offset, sect[data].sh_size) == -1) {
      perror(outfile);
      goto error;
    }

  /* Save text_offset to file header */
  lseek(outfd, 12, SEEK_SET);
  write(outfd, &text_offset, sizeof(int));
  close(outfd);

  /* Display segment sizes */
  if(verbose) {
    printf("Text Segment = %d\n", sect[text].sh_size);
    if(data != -1 && sect[data].sh_size)
      printf("Data Segment = %d\n", sect[data].sh_size);
    if(bss != -1 && sect[bss].sh_size)
      printf("Bss Segment = %d\n", sect[bss].sh_size);
    putchar('\n');
  }

  return 0;
}

// Process a 64-bit ELF File.
//
static int handle_64bit(ELF64 *head, unsigned char *input, int size)
{
  SECT64 *sect;
  RELA64 *rel;
  SYM64 *sym=NULL;
  struct stat statbuf;
  struct reloc reloc;
  struct symlist *undef_list=NULL, *ptr, *last;
  int text=-1, data=-1, bss=-1;
  int seg, type, link, strlink, text_offset;
  int i, j, n, outfd, error=0;
  char *strtab=NULL, *name;
  quad addend;

  /* Validate section table */
  sect=(SECT64 *)(input+head->e_shoff);
  for(i=0;i < head->e_shnum;i++) {
    if(sect[i].sh_type != SHT_NOBITS) {
      if(sect[i].sh_offset > size || sect[i].sh_offset < 0 ||
         sect[i].sh_offset+sect[i].sh_size > size || sect[i].sh_offset+sect[i].sh_size < 0) {
        fprintf(stderr, "%s: Section %d has invalid size\n", file, i);
        return 1;
      } if(sect[i].sh_size >= 0x8000000) {
        fprintf(stderr, "%s: Section %d is larger than 128 MB\n", file, i);
        return 1;
      }
    }

    /* Determine file offset for String Table */
    if(!strtab && sect[i].sh_type == SHT_STRTAB)
      strtab=input+sect[i].sh_offset;
  }
  if(!strtab) {
    fprintf(stderr, "%s: No string table present\n", file);
    return 1;
  }

  /* Print sections */
  if(verbose) {
    printf("Sections in %s:\n\n", file);
    printf("Sect: Name                Offset            Size              Info Link\n");
    for(i=0;i < head->e_shnum;i++)
      if(sect[i].sh_type != SHT_NULL)
        printf("%4d: %-18.18s  %016qx  %016qx  %-4d %-4d\n", i, strtab+sect[i].sh_name,
               (unsigned quad)sect[i].sh_offset, (unsigned quad)sect[i].sh_size,
               sect[i].sh_info, sect[i].sh_link);
  }


  /* Open output file */
  if((outfd=open(outfile, O_WRONLY|O_CREAT|O_TRUNC, 0666)) == -1) {
    perror(outfile);
    return 1;
  }

  /* Write .text .data .bss segment sizes */
  for(i=0,seg=0;i < head->e_shnum;i++)
    if(!strcmp(strtab+sect[i].sh_name, ".text")) {
      text=i;
      seg=sect[i].sh_size;
      break;
    }
  if(i == head->e_shnum) {
    fprintf(stderr, "%s: Missing .text section\n", file);
    goto error;
  }
  write(outfd, &seg, sizeof(int));

  for(i=0,seg=0;i < head->e_shnum;i++)
    if(!strcmp(strtab+sect[i].sh_name, ".data")) {
      data=i;
      seg=sect[i].sh_size;
      break;
    }
  write(outfd, &seg, sizeof(int));

  for(i=0,seg=0;i < head->e_shnum;i++)
    if(!strcmp(strtab+sect[i].sh_name, ".bss")) {
      bss=i;
      seg=sect[i].sh_size;
      break;
    }
  write(outfd, &seg, sizeof(int));

  /* Segment text offset in output file (Write 0 for now, fill in later) */
  seg=0;
  write(outfd, &seg, sizeof(int));


  /* Process Symbol Relocations */
  for(i=0;i < head->e_shnum;i++) {
    if(sect[i].sh_type != SHT_RELA && sect[i].sh_type != SHT_REL)
      continue;
    if((seg=sect[i].sh_info) < 0 || seg >= head->e_shnum || sect[seg].sh_type != SHT_PROGBITS)
      continue;
    if((link=sect[i].sh_link) < 0 || link >= head->e_shnum || sect[link].sh_type != SHT_SYMTAB)
      continue;
    if((strlink=sect[link].sh_link) < 0 || strlink >= head->e_shnum ||
       sect[strlink].sh_type != SHT_STRTAB)
      continue;

    /* Loop on # symbols per section */
    for(n=0;n < sect[i].sh_size/sizeof(RELA64);n++) {
      if(sect[i].sh_type == SHT_REL) {
        rel=(RELA64 *)((REL64 *)(input+sect[i].sh_offset)+n);
        addend=0;
      } else {
        rel=(RELA64 *)(input+sect[i].sh_offset)+n;
        addend=rel->r_addend;
      }
      type=Type64(rel->r_info);
      j=Sym64(rel->r_info);
      if((type != 1 && type != 2 && type != 10 && type != 11) || j < 0 ||
         j >= sect[link].sh_size/sizeof(SYM64) || rel->r_offset > sect[seg].sh_size) {
        fprintf(stderr, "%s: Invalid relocation type %d at Section %d, Entry %d\n", file,
                type, i, n);
        goto error;
      }

      /* Determine symbol name */
      sym=(SYM64 *)(input+sect[link].sh_offset)+j;
      if(sym->st_info == STT_SECTION)
        name=(sym->st_shndx < head->e_shnum) ?
             strtab+sect[sym->st_shndx].sh_name:"";
      else {
        if(sect[strlink].sh_offset+sym->st_name > size) {
          fprintf(stderr, "%s: Error retrieving symbol name for Section %d, Entry %d\n",
                  file, i, n);
          goto error;
        }
        name=input+sect[strlink].sh_offset+sym->st_name;
      }

      /* Display relocation list */
      if(verbose > 1) {
        if(!n)
          printf("\nSection %s relocations:\n", strtab+sect[seg].sh_name);
        printf("%4d: %-18.18s  %016qx  %5s:%016qx  %s\n", n, name, (unsigned quad)rel->r_offset,
               !sym->st_shndx?"UNDEF":sym->st_shndx < head->e_shnum ?
               strtab+sect[sym->st_shndx].sh_name:sym->st_shndx == SHN_ABS?"ABS":
               sym->st_shndx == SHN_COMMON?"COM":"?", (unsigned quad)sym->st_value+addend,
               (type == 10 || type == 11)?"32 Absolute":type == 2?"32 PC-Rel.":"");
      }

      /* Record a list of undefined symbols */
      if(!sym->st_shndx) {
        for(ptr=undef_list,last=NULL;ptr;last=ptr,ptr=ptr->next)
          if(!strcmp(ptr->name, name))
            break;
        if(!ptr) {
          ptr=malloc(sizeof(struct symlist));
          ptr->next=NULL;
          ptr->name=strdup(name);
          if(last)
            last->next=ptr;
          else
            undef_list=ptr;
        }
        continue;
      }

      /* If this is an absolute address that is known, modify segment directly */
      if(sym->st_shndx == SHN_ABS && (type == 10 || type == 11)) {
        *(int *)(input+sect[seg].sh_offset+rel->r_offset) += sym->st_value+addend;
        continue;  /* 32-bit absolute */
      }
      if(sym->st_shndx == SHN_ABS && type == 1) {
        *(quad *)(input+sect[seg].sh_offset+rel->r_offset) += sym->st_value+addend;
        continue;  /* 64-bit absolute */
      }

      /* If this symbol is Common, delay error reporting */
      if(sym->st_shndx == SHN_COMMON) {
        error=1;
        continue;
      }

      /* Otherwise store in output module file */
      reloc.src=(seg == data);
      reloc.mode=(type == 10 || type == 11)?0:(type == 2)?1:2;
      reloc.dest=(sym->st_shndx == text)?1:(sym->st_shndx == data)?2:(sym->st_shndx == bss)?3:0;
      reloc.offset=rel->r_offset;
      reloc.value=sym->st_value+addend;
      write(outfd, &reloc, (type == 1)?12:8);
    }
  }

  if(verbose > 1)
    putchar('\n');

  /* Linker encountered an error */
  if(undef_list || error) {
    /* List undefined symbols */
    for(ptr=undef_list;ptr;ptr=ptr->next)
      fprintf(stderr, "%s: Error: symbol `%s' is undefined\n", file, ptr->name);

    /* Report use of Common variables */
    if(error) {
      fprintf(stderr, "%s: Error: At least one symbol was defined as Common.\n", file);
      fprintf(stderr, "%s: Error: Recompile the program using -fno-common.\n", file);
    }

    error:

    /* Remove temporary file on error */
    fstat(outfd, &statbuf);
    if(S_ISREG(statbuf.st_mode))
      unlink(outfile);

    close(outfd);
    return 1;
  }

  /* Write text and data segments to file */
  text_offset=lseek(outfd, 0, SEEK_CUR);
  if(write(outfd, input+sect[text].sh_offset, sect[text].sh_size) == -1) {
    perror(outfile);
    goto error;
  }
  if(data != -1 && sect[data].sh_size)
    if(write(outfd, input+sect[data].sh_offset, sect[data].sh_size) == -1) {
      perror(outfile);
      goto error;
    }

  /* Save text_offset to file header */
  lseek(outfd, 12, SEEK_SET);
  write(outfd, &text_offset, sizeof(int));
  close(outfd);

  /* Display segment sizes */
  if(verbose) {
    printf("Text Segment = %qd\n", (unsigned quad)sect[text].sh_size);
    if(data != -1 && sect[data].sh_size)
      printf("Data Segment = %qd\n", (unsigned quad)sect[data].sh_size);
    if(bss != -1 && sect[bss].sh_size)
      printf("Bss Segment = %qd\n", (unsigned quad)sect[bss].sh_size);
    putchar('\n');
  }

  return 0;
}

// Validates an ELF file header. Returns 0 if the header is corrupt or unreadable. Returns 1 if the
// file is 32-bit, or 2 if the file is 64-bit.
//
static int elf_header(ELF32 *elf, char *file, int size)
{
  ELF64 *elf64=(ELF64 *)elf;
  unsigned char *data=(char *)elf;

  if(*(int *)elf != 0x464c457f) {
    fprintf(stderr, "%s: Not an ELF object file\n", file);
    return 0;
  }

  if(data[EI_CLASS] != ELFCLASS32 && data[EI_CLASS] != ELFCLASS64) {
    fprintf(stderr, "%s: Invalid ELF class\n", file);
    return 0;
  }

  if(data[EI_DATA] != ELFDATA2LSB) {
    fprintf(stderr, "%s: Invalid data encoding\n", file);
    return 0;
  }

  if(data[EI_VERSION] != EV_CURRENT) {
    fprintf(stderr, "%s: Invalid ELF version\n", file);
    return 0;
  }

  if(elf->e_type != ET_REL) {
    fprintf(stderr, "%s: Not a relocatable file\n", file);
    return 0;
  }

  if(elf->e_machine != EM_386 && elf->e_machine != EM_X86_64) {
    fprintf(stderr, "%s: Unsupported ELF machine type\n", file);
    return 0;
  }

  if(data[EI_CLASS] == ELFCLASS64) {
    /* 64-bit ELF header */
    if(elf64->e_phoff || elf64->e_phentsize || elf64->e_phnum) {
      fprintf(stderr, "%s: Cannot link ELF object containing program headers\n", file);
      return 0;
    }

    if(elf64->e_shentsize != sizeof(SECT64)) {
      fprintf(stderr, "%s: Invalid section entry size\n", file);
      return 0;
    }

    if(elf64->e_shoff + elf64->e_shentsize*elf64->e_shnum > size) {
      fprintf(stderr, "%s: Cannot read truncated ELF file\n", file);
      return 0;
    }
    return 2;
  }

  /* 32-bit ELF header */
  if(elf->e_phoff || elf->e_phentsize || elf->e_phnum) {
    fprintf(stderr, "%s: Cannot link ELF object containing program headers\n", file);
    return 0;
  }

  if(elf->e_shentsize != sizeof(SECT32)) {
    fprintf(stderr, "%s: Invalid section entry size\n", file);
    return 0;
  }

  if(elf->e_shoff + elf->e_shentsize*elf->e_shnum > size) {
    fprintf(stderr, "%s: Cannot read truncated ELF file\n", file);
    return 0;
  }

  return 1;
}
