/* mopen.c - Dynamically loads a Module File (.m) in memory */

// Written by Byron Stanoszek, Sep 18 2003  (gandalf@winds.org)
// vim:tw=99

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/mman.h>

#define quad long long

/* 64-bit or 96-bit relocation record */
struct reloc {
  unsigned int offset:27;  // Source Section Offset
  unsigned int dest:2;     // Destination Section (0=Absolute, 1=.text, 2=.data, 3=.bss)
  unsigned int mode:2;     // Addressing Mode (0=32 Absolute, 1=32 PC-Relative, 2=64 Absolute)
  unsigned int src:1;      // Source Section (0=.text, 1=.data)

  unsigned quad value;     // Relocation Value
} __attribute__((__packed__));

/* Structure containing array of free blocks in memory */
static struct heap {
  int offset;   // Offset from byte 0
  int free;     // Amount of free space available (0=EOF)
} *heap;

static int heap_size;  // Number of array entries in 'heap'.

static void *memory;   // Memory address of start of heap (mmap'd)

// Allocates <len> bytes out of the cache file, and returns a pointer to the new data in memory.
//
static void *heap_alloc(int len)
{
  int i, offset;

  if(!len)
    return NULL;

  /* Align to multiples of 16 bytes */
  len=(len+15) & ~15;

  /* Start a new heap */
  if(!heap) {
    heap=malloc(sizeof(struct heap));
    heap[0].offset=len;
    heap[0].free=0;
    heap_size=1;
    memory=mmap((void *)0x1000000, len, PROT_READ|PROT_WRITE|PROT_EXEC, MAP_PRIVATE|MAP_ANONYMOUS,
                0, 0);
    return memory;
  }

  /* Find a free part of the heap that's big enough to hold our request */
  for(i=0;heap[i].free;i++)
    if(heap[i].free >= len) {
      offset=heap[i].offset;
      heap[i].offset += len;
      heap[i].free -= len;

      /* Shrink heap array if we allocate the entire block */
      if(!heap[i].free) {
        heap_size--;
        memmove(heap+i, heap+i+1, (heap_size-i)*sizeof(struct heap));
        heap=realloc(heap, heap_size * sizeof(struct heap));
      }
      return memory+offset;
    }

  /* Append data to end of file */
  offset=heap[i].offset;
  heap[i].offset += len;
  mremap(memory, offset, heap[i].offset, 0);

  return memory+offset;
}

// Frees a chunk of data of size <len> starting at <start> in heap memory. It consolidates
// neighboring free blocks to minimize fragmentation.
//
__attribute__((__unused__))
static void heap_free(void *start, int len)
{
  int i, offset=start-memory;

  if(!heap || !len)
    return;

  /* Align to multiples of 16 bytes */
  len=(len+15) & ~15;

  /* Scan heap searching for free neighbor blocks */
  for(i=0;heap[i].free;i++) {
    /* Is existing free-block after our deleted one? */
    if(heap[i].offset-len == offset) {
      heap[i].offset -= len;
      heap[i].free += len;
      return;
    }

    /* Is existing free-block before our deleted one? */
    if(heap[i].offset+heap[i].free == offset) {
      heap[i].free += len;

      /* Consolidate if there's another free-block after ours */
      if(heap[i].offset+heap[i].free == heap[i+1].offset) {
        if(!heap[i+1].free) {
          mremap(memory, heap[i+1].offset, heap[i].offset, 0);
          heap_size--;
          heap=realloc(heap, heap_size * sizeof(struct heap));
          heap[i].free=0;
          return;
        }
        heap[i].free += heap[i+1].free;
        i++;
        heap_size--;
        memmove(heap+i, heap+i+1, (heap_size-i)*sizeof(struct heap));
        heap=realloc(heap, heap_size * sizeof(struct heap));
      }
      return;
    }
  }

  /* Determine if we're freeing the last block in the file */
  if(heap[i].offset-len == offset) {
    mremap(memory, heap[i].offset, heap[i].offset-len, 0);
    heap[i].offset -= len;
    return;
  }

  /* Insert new free area into heap array */
  for(i=0;heap[i].offset < offset;i++);
  heap_size++;
  heap=realloc(heap, heap_size * sizeof(struct heap));
  memmove(heap+i+1, heap+i, (heap_size-i-1)*sizeof(struct heap));
  heap[i].offset=offset;
  heap[i].free=len;
}

// Primary function to load a module in memory. Returns a pointer to the 'text' segment.
// If 'data', or 'bss' is non-null, also stores pointers to those relative segments.
//
// Returns NULL on error, with 'errno' set to the appropriate error, with 'data' and 'bss' left
// unchanged.
//
void *mopen(char *file, void **data, void **bss)
{
  struct reloc *rel;
  struct stat statbuf;
  char *input=NULL, *memory, *src, *s;
  int *offset, sz_text, sz_data, sz_bss, total;
  int temp_errno, fd, size;

  /* Open specified file */
  if((fd=open(file, O_RDONLY)) == -1)
    return NULL;

  /* Find and validate file size */
  if(lstat(file, &statbuf)) {
    temp_errno=errno;
    close(fd);
    errno=temp_errno;
    return NULL;
  } if(statbuf.st_size < 16) {
    close(fd);
    errno=ELIBBAD;
    return NULL;
  } if(statbuf.st_size > 0xfffffff+16) {
    close(fd);
    errno=EFBIG;
    return NULL;
  }
  size=statbuf.st_size;

  /* Map file into memory */
  if(!(input=mmap(NULL, size, PROT_READ, MAP_SHARED, fd, 0))) {
    temp_errno=errno;
    close(fd);
    errno=temp_errno;
    return NULL;
  }
  offset=(int *)input;

  /* Calculate segment size required */
  sz_text=offset[0]+(~(offset[0]-1) & 15);
  sz_data=offset[1]+(~(offset[1]-1) & 15);
  sz_bss=offset[2]+(~(offset[2]-1) & 15);
  total=sz_text+sz_data+sz_bss;
  if(offset[3] < 0 || offset[3] > size || offset[0]+offset[1] != size-offset[3]) {
    munmap(input, size);
    close(fd);
    errno=ELIBBAD;
    return NULL;
  }

  /* Allocate the new memory */
  memory=heap_alloc(total);

  /* Copy text and data segments */
  memcpy(memory, input+offset[3], offset[0]);
  memcpy(memory+sz_text, input+offset[3]+offset[0], offset[1]);
  memset(memory+sz_text+sz_data, 0, sz_bss);

  /* Handle relocations */
  s=input+16;
  while(s < input+offset[3]) {
    rel=(struct reloc *)s;
    src=memory+rel->offset+(rel->src?sz_text:0);
    if(rel->mode == 2) {
      /* 64-bit value */
      if(src >= memory && src <= memory+total-8)
        *(quad *)src += rel->value + (!rel->dest?0:(rel->dest == 1)?(long)memory:(rel->dest == 2)?
                         ((long)memory+sz_text):((long)memory+sz_text+sz_data));
      s+=12;
    } else {
      /* 32-bit value */
      if(src >= memory && src <= memory+total-4)
        *(int *)src += (int)rel->value + (!rel->dest?0:(rel->dest == 1)?(long)memory:
                        (rel->dest == 2)?((long)memory+sz_text):
                        ((long)memory+sz_text+sz_data)) - (rel->mode?(long)src:0);
      s+=8;
    }
  }

  munmap(input, size);
  close(fd);

  /* Set output pointers */
  if(data)
    *data=memory+sz_text;
  if(bss)
    *bss=memory+sz_text+sz_data;

  return memory;
}
